/*
 * FILE: STunnel.c
 * AUTH: Soren Spies (sspies)
 * DATE: 5 Feb 2001
 * DESC: test driver for STunnel project
 *
 * Copyright (c) 2001 Soren Spies.
 * Please see included file COPYING for license details.
 * STunnel is available from http://www.techhouse.org/~soren/software
 */

/*#define DEBUG */

#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <netinet/in.h>

#include <arpa/inet.h>
#include <sys/socket.h>
#include <signal.h>
#include <stdio.h>
#include <sys/wait.h>
#include <unistd.h>

#include <util.h>

#include "util.h"		/* add ourhdr.h stuff? */
#include "ST.h"

STReturn_t
STDeleteKey(char *delhost)
{
	FILE *keyf, *tkeyf;
	char kfpath[PATH_MAX], tkfpath[PATH_MAX];
	char line[LINECHARS];

	snprintf(kfpath, PATH_MAX, KFFMT, getenv("HOME"), KNOWNHOSTSSUFFIX);
	if(!(keyf = fopen(kfpath, "r")))
		reterr("couldn't open known-hosts file");
	snprintf(tkfpath, PATH_MAX, TKFFMT, kfpath, getpid());
	if(!(tkeyf = fopen(tkfpath, "w")))
		reterr("couldn't open temp file");

	dprintf(("deleting %s from %s using %s\n", delhost, kfpath, tkfpath));
	while(fgets(line, LINECHARS, keyf)) {
		/*dprintf(("line: %40s", line));*/
		if(!strstr(line, delhost)) {
			dprintf(("accepted a line\n"));
			if(fputs(line, tkeyf)) {
				dprintf(("fputs failed, but reterr didn't say anything?\n"));
				unlink(tkfpath);
				reterr("fputs failed");
			}
		}
	}

	if(rename(tkfpath, kfpath))
		reterr("couldn't rename temp file to real file");

	dprintf(("returning success\n"));
	return(STSuccess);
}

STReturn_t
STOpenTunnel(struct sockaddr_in *src, struct sockaddr_in *dst, STOpts opts, STConRef *ref)
{
	char *args[NUMARGS], ports[ARGSZ], userhost[ARGSZ], pts_name[PATH_MAX];
	/*int sv[2];*/
	STConRef conref;

	/* set up arguments */
	args[0] = SSH;
	if(src->sin_addr.s_addr == INADDR_LOOPBACK) {
		/* ssh -L */
		args[1] = LOCALBIND;
		snprintf(userhost, ARGSZ, USERHOSTFMT, opts.login, 
				inet_ntoa(dst->sin_addr));
	}
	else {
		/* ssh -R */
		if(dst->sin_addr.s_addr != INADDR_LOOPBACK)
			reterr("need LOOPBACK in either source or destination addr");
		args[1] = REMOTEBIND;
		snprintf(userhost, ARGSZ, USERHOSTFMT, opts.login, 
				inet_ntoa(src->sin_addr));
	}

	/* this is the same either way ... always src -> dst */
	snprintf(ports, ARGSZ, PORTFMT, src->sin_port,
			inet_ntoa(dst->sin_addr), dst->sin_port);
	dprintf(("ports: %s\n", ports));
	args[2] = ports;

	args[3] = userhost;
	args[4] = ONEATTEMPTOPT;
	args[5] = ONEPROMPTOPT;
	args[6] = NORSHOPT;
	args[7] = SSHSHELLCMD;
	args[8] = NULL;

	dprintf(("running %s %s %s %s\n", args[0], args[1], args[2], 
			args[3], args[4]));

	/* args checked, allocate a reference for this connection */
	conref = (STConRef)malloc(sizeof(struct STConData));
	*ref = conref;	/* now caller's ptr is the same as ours */
	conref->state = Open;

	/*if(socketpair(AF_UNIX, SOCK_STREAM, 0, sv))
		reterr("socketpair failed\n");*/
	dprintf(("forking ...\n"));

	if(conref->pid = forkpty(&conref->pipe, pts_name, NULL, NULL)) {
		/* parent */
		dprintf(("child pid = %d\ttty = %s\n", conref->pid, pts_name));
		if(conref->pid < 0) {
			free(conref);
			reterr("could not fork?");
		}
		return STContOpen(conref, NULL);
	}
	else {	/* child */
		FILE *f = fopen("/tmp/bob", "w+");
		struct termios buf;

		fprintf(f, "hi, I am %d\n", getpid());
		fflush(f);

		if(tcgetattr(STDIN_FILENO, &buf) < 0)
			reterr("couldn't get old terminal characteristics");
		cfmakeraw(&buf);
		if(tcsetattr(STDIN_FILENO, TCSANOW, &buf) < 0)
			reterr("couldn't set terminal characteristics");

		/*
		setsid();
		close(sv[0]);
		dup2(sv[1], STDIN_FILENO);
		dup2(sv[1], STDOUT_FILENO);
		dup2(sv[1], STDERR_FILENO);
		*/
		execv(args[0], args);
		reterr("execv failed");
	}
}

STReturn_t
STContOpen(STConRef conref, STContParams *params)
{
	char line[LINECHARS];
	int nb;

	dprintf(("STContOpen called\n"));
	if(!params && conref->state != Open)
		reterr("I need params unless I'm called by STOpenTunnel");

	switch(conref->state) {
		case Closed:
			reterr("don't ask me to open closed connections");

		case Open:
			// I was probably called by STOpenTunnel(.)
			break;

		case NewKey:
			if(!params->goahead)
				reterr("you should just call STCloseTunnel");
			dprintf(("writing %s to the pipe\n", YESSTR));
			write(conref->pipe, YESSTR, sizeof(YESSTR) - 1);
			break;

		case PPNeeded:
			dprintf(("parent writing passphrase|word to pipe\n"));
			write(conref->pipe, params->passphrase, strlen(params->passphrase));
			write(conref->pipe, NEWLINE, sizeof(NEWLINE) - 1);
			break;

		case Connected:
			reterr("connection is already connected");
	}

	bzero(line, LINECHARS);
	while(nb = read(conref->pipe, line, LINECHARS)) {
		char *l = line;

		line[nb] = NULL;
		dprintf(("line from child (%d):\n%sXXX\n", nb, line));
		if(line[0] == '\n' && line[1] == '\0')
			continue;
		while(*l == '\n' || *l == '\r' || *l == '@')
			l++;
		if(!*l)
			continue;
		if(strstr(l, NEWKEYSTR)) {
			conref->state = NewKey;
			return STNewHostKey;
		}
		if(strstr(l, PWSTR)) {
			conref->state = PPNeeded;
			return STPasswordNeeded;
		}
		if(strstr(l, PPSTR)) {
			conref->state = PPNeeded;
			return STPassphraseNeeded;
		}
		if(strstr(l, KEYCHANGEDSTR)) {
			kill(conref->pid, SIGQUIT);
			STCloseTunnel(conref);
			return STHostKeyChanged;
		}
		if(strstr(l, PERMDENIED)) {
			STCloseTunnel(conref);
			return STPermissionDenied;
		}
		if(strstr(l, BADPPHRASE))
			continue;
		/*if(strstr(l, UNKNOWNHOST)) {
			STCloseTunnel(conref);
			return STUnknownHost;
		}*/
		// BOBSTR means we've connected
		if(!strcmp(l, BOBSTR)) {
			conref->state = Connected;
			return STSuccess;
		}

		/* if there is nothing but warnings, then we might as well try more */
		if(strstr(l, WARNING)) {
			continue;
		}

		// we really shouldn't get here
		STCloseTunnel(conref);
		dprintf(("falling off the end of the searches\n"));
		reterr("ssh didn't respond as we expected");
	}
	STCloseTunnel(conref);
	dprintf(("falling off the end of the searches\n"));
	reterr("ssh didn't respond as we expected");
}

STReturn_t
STCloseTunnel(STConRef conref)
{
	int status;

	dprintf(("STCloseTunnel called on %p\n", conref));
	if(kill(conref->pid, SIGHUP) < 0)
		dperror("kill -HUP had trouble");
	if(waitpid(conref->pid, &status, WNOHANG) < 0)
		dperror(("1st waitpid had trouble"));
	if(!WIFEXITED(status) && !WIFSIGNALED(status)) {
		dprintf(("SIGHUP did not do the job\n"));
		kill(conref->pid, SIGKILL);
		if(waitpid(conref->pid, &status, WNOHANG) < 0)
			dperror(("2nd waitpid had trouble"));
	}

	free(conref);
	return STSuccess;
}

/* OS X already has forkpty? */
#if 0	/* ignore the rest of the file */
int
tty_raw(int fd)

int
ptym_open(char *pts_name)
{
	int fdm;
	char *ptr1, *ptr2;

	strcpy(pts_name, "/dev/ptyXY");
		/* array index:  01234567890 ? */
	for(ptr1 = "pqrstuvwxyzPQRST"; *ptr1 != 0; ptr1++) {
		pts_name[8] = *ptr1;
		for(ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++) {
			pts_name[9] = *ptr2;

			if((fdm = open(pts_name, O_RDWR)) < 0) {
				if(errno == ENOENT)
					return(-1);		/* out of pty's */
				else
					continue		/* try next pty */
			}
			pts_name[5] = 't';		/* "pty" -> "tty"
			return(fdm);			/* return fd of master */
		}
	}
	return(-1);		/* failure -- no more to try */
}

int
pts_open(int fdm, char *pts_name)
{
	int fds;
	/* this is all root-only stuff
	 * struct group *grouptr;
	 * int gid, fds;

	 * if((grptr = getgrname("tty")) != NULL)
		gid = grptr->gr_gid;
	 * else
		gid = -1;		/* group 'tty' doesn't exist? */

	 * stuff root could do
	 * chown(pts_name, getuid(), gid);
	 * chmod(pts_name, S_IRUSR|S_IWUSR|S_IWGRP); */

	 if((fds = open(pts_name, O_RDWR)) < 0) {
		close(fdm);
		return(-1);		/* couldn't open slave? */
	}
	return fds;			/* slave tty */
}

pid_t pty_fork(int ptrfdm, char *slave_name,
		const struct termios *slave_termios,
		const struct winsize *slave_winsize)
{
#endif
