/*
 * FILE: STunnel.h
 * AUTH: Soren Spies (sspies)
 * DATE: 5 Feb 2001
 * 
 * STunnel is a package for easily setting up SSH tunnels using 
 * existing command line tools.  The SSH protocol provides encrypted 
 * connections using public key techniques.  OpenSSH is a freely-available 
 * implementation of this protocol and is the package with which STunnel 
 * has been tested.  The STunnel package is being released under a the 
 * BSD license so that it can be useful to the greatest number of people.
 *
 * See the included file README for more instructions.
 *
 * Copyright (c) 2001 Soren Spies.
 * Please see included file COPYING for license details.
 * STunnel is available from http://www.techhouse.org/~soren/software
 *
 * Public Functions:
 * STOpenTunnel(.) - begin opening a tunnel
 * STContOpen(.) - continue with further info such as a password
 * STCloseTunnel(.) - close a tunnel opened by OpenTunnel
 * STCancelOpening(.) - (same as CloseTunnel(.)?)
 * STDeleteKey(.) - delete a key that is no longer valid
 */

/* Public interface */
struct STConData;
struct sockaddr_in;
typedef struct STConData* STConRef;		/* ST connection reference */

/* return types for STOpenTunnel(.) and STContOpen(.)
 * if you don't want continue, call STCloseTunnel(.) */
typedef enum STReturn {
	STSuccess = 0,
	STNewHostKey,		/* 1 */
	STPasswordNeeded, 	/* 2 */
	STPassphraseNeeded,	/* 3 */
	STHostKeyChanged,	/* 4 */
	STErr,				/* 5 */
	STPermissionDenied,	/* 6 */
	/*STUnknownHost,	// you have to pass us an IP */
	STTimeout			/* 8 */
} STReturn_t;

/* options that can be passed to SSH login is required */
typedef struct STOpts {
	char *login;		/* username for remote system */
	char **ssh_opts;	/* not checked, just passed along */
} STOpts;

/* stuff you might need to pass to STContOpen(.) */
typedef struct STContParams {
	char *passphrase;	/* for password or passphrase */
	int goahead;		/* accept new key, etc - non-zero means yes */
} STContParams;

/* 
 * STOpenTunnel
 * src and dst specify hostname / ports for each end of tunnel.
 * One end should be set to INADDR_LOOPBACK to indicate the side
 * of the tunnel that should be local.  Further calls to STContOpen(.)
 * will be necessary unless STSuccess is returned.
 */
STReturn_t
STOpenTunnel(struct sockaddr_in *src, struct sockaddr_in *dst, STOpts opts, STConRef *conref);

/*
 * STContOpen
 * continue open with further parameters such as password
 */
STReturn_t
STContOpen(STConRef, STContParams*);

/*
 * STCloseTunnel
 * closes tunnel and frees any resources ... invalidates given connection ref
 */
STReturn_t
STCloseTunnel(STConRef);

/*
 * STDeleteKey
 * remove all mention of the given host from $HOME/.ssh/known_hosts
 */
STReturn_t
STDeleteKey(char *host);
